@@ -1,68 +1,80 @@
-use core::convert::TryFrom;
-
 use serde::de::{self, DeserializeSeed, EnumAccess, MapAccess, SeqAccess, Visitor};
 use static_assertions::assert_impl_all;
 
-use std::{ffi::CStr, marker::PhantomData, os::unix::io::RawFd, str};
+use std::{ffi::CStr, marker::PhantomData, str};
+
+#[cfg(unix)]
+use std::os::fd::AsFd;
 
 use crate::{
-    de::ValueParseStage, framing_offset_size::FramingOffsetSize, framing_offsets::FramingOffsets,
-    signature_parser::SignatureParser, utils::*, EncodingContext, EncodingFormat, Error, Result,
-    Signature,
+    de::{DeserializerCommon, ValueParseStage},
+    framing_offset_size::FramingOffsetSize,
+    framing_offsets::FramingOffsets,
+    serialized::{Context, Format},
+    signature_parser::SignatureParser,
+    utils::*,
+    Basic, Error, Result, Signature,
 };
 
 /// Our GVariant deserialization implementation.
 #[derive(Debug)]
-pub struct Deserializer<'de, 'sig, 'f, B>(pub(crate) crate::DeserializerCommon<'de, 'sig, 'f, B>);
+pub struct Deserializer<'de, 'sig, 'f, F>(pub(crate) DeserializerCommon<'de, 'sig, 'f, F>);
 
-assert_impl_all!(Deserializer<'_, '_,'_, i32>: Send, Sync, Unpin);
+assert_impl_all!(Deserializer<'_, '_,'_, ()>: Send, Sync, Unpin);
 
-impl<'de, 'sig, 'f, B> Deserializer<'de, 'sig, 'f, B>
-where
-    B: byteorder::ByteOrder,
-{
+impl<'de, 'sig, 'f, F> Deserializer<'de, 'sig, 'f, F> {
     /// Create a Deserializer struct instance.
-    pub fn new<'r: 'de>(
+    ///
+    /// On Windows, the function doesn't have `fds` argument.
+    pub fn new<'r: 'de, S>(
         bytes: &'r [u8],
-        fds: Option<&'f [RawFd]>,
-        signature: &Signature<'sig>,
-        ctxt: EncodingContext<B>,
-    ) -> Self {
-        assert_eq!(ctxt.format(), EncodingFormat::GVariant);
+        #[cfg(unix)] fds: Option<&'f [F]>,
+        signature: S,
+        ctxt: Context,
+    ) -> Result<Self>
+    where
+        S: TryInto<Signature<'sig>>,
+        S::Error: Into<Error>,
+    {
+        assert_eq!(ctxt.format(), Format::GVariant);
 
-        let sig_parser = SignatureParser::new(signature.clone());
-        Self(crate::DeserializerCommon {
+        let signature = signature.try_into().map_err(Into::into)?;
+        let sig_parser = SignatureParser::new(signature);
+        Ok(Self(DeserializerCommon {
             ctxt,
             sig_parser,
             bytes,
+            #[cfg(unix)]
             fds,
+            #[cfg(not(unix))]
+            fds: PhantomData,
             pos: 0,
-            b: PhantomData,
-        })
+            container_depths: Default::default(),
+        }))
     }
 }
 
 macro_rules! deserialize_basic {
     ($method:ident) => {
-        #[inline]
         fn $method<V>(self, visitor: V) -> Result<V::Value>
         where
             V: Visitor<'de>,
         {
-            let ctxt = EncodingContext::new_dbus(self.0.ctxt.position() + self.0.pos);
+            let ctxt = Context::new_dbus(self.0.ctxt.endian(), self.0.ctxt.position() + self.0.pos);
 
-            let mut dbus_de = crate::dbus::Deserializer::<B>(crate::DeserializerCommon::<B> {
+            let mut dbus_de = crate::dbus::Deserializer::<F>(DeserializerCommon::<F> {
                 ctxt,
                 sig_parser: self.0.sig_parser.clone(),
-                bytes: &self.0.bytes[self.0.pos..],
+                bytes: subslice(self.0.bytes, self.0.pos..)?,
                 fds: self.0.fds,
                 pos: 0,
-                b: PhantomData,
+                container_depths: self.0.container_depths,
             });
 
             let v = dbus_de.$method(visitor)?;
             self.0.sig_parser = dbus_de.0.sig_parser;
             self.0.pos += dbus_de.0.pos;
+            // Basic types don't have anything to do with container depths so not updating it here.
 
             Ok(v)
         }
@@ -84,9 +96,8 @@ macro_rules! deserialize_as {
     }
 }
 
-impl<'de, 'd, 'sig, 'f, B> de::Deserializer<'de> for &'d mut Deserializer<'de, 'sig, 'f, B>
-where
-    B: byteorder::ByteOrder,
+impl<'de, 'd, 'sig, 'f, #[cfg(unix)] F: AsFd, #[cfg(not(unix))] F> de::Deserializer<'de>
+    for &'d mut Deserializer<'de, 'sig, 'f, F>
 {
     type Error = Error;
 
@@ -94,9 +105,9 @@ where
     where
         V: Visitor<'de>,
     {
-        let c = self.0.sig_parser.next_char();
+        let c = self.0.sig_parser.next_char()?;
 
-        crate::de::deserialize_any::<B, Self, V>(self, c, visitor)
+        crate::de::deserialize_any::<Self, V>(self, c, visitor)
     }
 
     deserialize_basic!(deserialize_bool);
@@ -140,9 +151,8 @@ where
     where
         V: Visitor<'de>,
     {
-        let s = if self.0.sig_parser.next_char() == VARIANT_SIGNATURE_CHAR {
-            let slice = &self.0.bytes[self.0.pos..];
-
+        let slice = subslice(self.0.bytes, self.0.pos..)?;
+        let s = if self.0.sig_parser.next_char()? == VARIANT_SIGNATURE_CHAR {
             if slice.contains(&0) {
                 return Err(serde::de::Error::invalid_value(
                     serde::de::Unexpected::Char('\0'),
@@ -153,14 +163,16 @@ where
             // GVariant decided to skip the trailing nul at the end of signature string
             str::from_utf8(slice).map_err(Error::Utf8)?
         } else {
-            let cstr =
-                CStr::from_bytes_with_nul(&self.0.bytes[self.0.pos..]).map_err(|_| -> Error {
+            let cstr = CStr::from_bytes_with_nul(slice).map_err(|_| -> Error {
+                let unexpected = if self.0.bytes.is_empty() {
+                    de::Unexpected::Other("end of byte stream")
+                } else {
                     let c = self.0.bytes[self.0.bytes.len() - 1] as char;
-                    de::Error::invalid_value(
-                        de::Unexpected::Char(c),
-                        &"nul byte expected at the end of strings",
-                    )
-                })?;
+                    de::Unexpected::Char(c)
+                };
+
+                de::Error::invalid_value(unexpected, &"nul byte expected at the end of strings")
+            })?;
             let s = cstr.to_str().map_err(Error::Utf8)?;
             self.0.pos += s.len() + 1; // string and trailing null byte
 
@@ -171,12 +183,21 @@ where
         visitor.visit_borrowed_str(s)
     }
 
+    #[cfg(feature = "option-as-array")]
+    fn deserialize_option<V>(self, _visitor: V) -> Result<V::Value>
+    where
+        V: Visitor<'de>,
+    {
+        panic!("`option-as-array` and `gvariant` features are incompatible. Don't enable both.");
+    }
+
+    #[cfg(not(feature = "option-as-array"))]
     fn deserialize_option<V>(self, visitor: V) -> Result<V::Value>
     where
         V: Visitor<'de>,
     {
         let signature = self.0.sig_parser.next_signature()?;
-        let alignment = alignment_for_signature(&signature, self.0.ctxt.format());
+        let alignment = alignment_for_signature(&signature, self.0.ctxt.format())?;
         let child_sig_parser = self.0.sig_parser.slice(1..);
         let child_signature = child_sig_parser.next_signature()?;
         let child_sig_len = child_signature.len();
@@ -191,28 +212,32 @@ where
 
             visitor.visit_none()
         } else {
-            let ctxt =
-                EncodingContext::new(self.0.ctxt.format(), self.0.ctxt.position() + self.0.pos);
+            let ctxt = Context::new(
+                self.0.ctxt.format(),
+                self.0.ctxt.endian(),
+                self.0.ctxt.position() + self.0.pos,
+            );
             let end = if fixed_sized_child {
                 self.0.bytes.len()
             } else {
                 self.0.bytes.len() - 1
             };
 
-            let mut de = Deserializer::<B>(crate::DeserializerCommon {
+            let mut de = Deserializer::<F>(DeserializerCommon {
                 ctxt,
                 sig_parser: self.0.sig_parser.clone(),
-                bytes: &self.0.bytes[self.0.pos..end],
+                bytes: subslice(self.0.bytes, self.0.pos..end)?,
                 fds: self.0.fds,
                 pos: 0,
-                b: PhantomData,
+                container_depths: self.0.container_depths.inc_maybe()?,
             });
 
             let v = visitor.visit_some(&mut de)?;
             self.0.pos += de.0.pos;
+            // No need for retaking the container depths as the underlying type can't be incomplete.
 
             if !fixed_sized_child {
-                let byte = self.0.bytes[self.0.pos];
+                let byte = *subslice(self.0.bytes, self.0.pos)?;
                 if byte != 0 {
                     return Err(de::Error::invalid_value(
                         de::Unexpected::Bytes(&byte.to_le_bytes()),
@@ -232,10 +257,10 @@ where
     where
         V: Visitor<'de>,
     {
-        let byte = self.0.bytes[self.0.pos];
+        let byte = *subslice(self.0.bytes, self.0.pos)?;
         if byte != 0 {
             return Err(de::Error::invalid_value(
-                de::Unexpected::Bytes(&self.0.bytes[self.0.pos..self.0.pos + 1]),
+                de::Unexpected::Bytes(&byte.to_le_bytes()),
                 &"0 byte expected for empty tuples (unit type)",
             ));
         }
@@ -263,7 +288,7 @@ where
     where
         V: Visitor<'de>,
     {
-        match self.0.sig_parser.next_char() {
+        match self.0.sig_parser.next_char()? {
             VARIANT_SIGNATURE_CHAR => {
                 self.0.sig_parser.skip_char()?;
                 self.0.parse_padding(VARIANT_ALIGNMENT_GVARIANT)?;
@@ -273,7 +298,7 @@ where
             }
             ARRAY_SIGNATURE_CHAR => {
                 self.0.sig_parser.skip_char()?;
-                let next_signature_char = self.0.sig_parser.next_char();
+                let next_signature_char = self.0.sig_parser.next_char()?;
                 let array_de = ArrayDeserializer::new(self)?;
 
                 if next_signature_char == DICT_ENTRY_SIG_START_CHAR {
@@ -284,7 +309,7 @@ where
             }
             STRUCT_SIG_START_CHAR => {
                 let signature = self.0.sig_parser.next_signature()?;
-                let alignment = alignment_for_signature(&signature, self.0.ctxt.format());
+                let alignment = alignment_for_signature(&signature, self.0.ctxt.format())?;
                 self.0.parse_padding(alignment)?;
 
                 self.0.sig_parser.skip_char()?;
@@ -292,19 +317,36 @@ where
                 let start = self.0.pos;
                 let end = self.0.bytes.len();
                 let offset_size = FramingOffsetSize::for_encoded_container(end - start);
-                visitor.visit_seq(StructureDeserializer {
+                self.0.container_depths = self.0.container_depths.inc_structure()?;
+                let v = visitor.visit_seq(StructureDeserializer {
                     de: self,
                     start,
                     end,
                     offsets_len: 0,
                     offset_size,
+                });
+                self.0.container_depths = self.0.container_depths.dec_structure();
+
+                v
+            }
+            <u8 as Basic>::SIGNATURE_CHAR => {
+                // Empty struct: encoded as a `0u8`.
+                let _: u8 = serde::Deserialize::deserialize(&mut *self)?;
+
+                let start = self.0.pos;
+                let end = self.0.bytes.len();
+                visitor.visit_seq(StructureDeserializer {
+                    de: self,
+                    start,
+                    end,
+                    offsets_len: 0,
+                    offset_size: FramingOffsetSize::U8,
                 })
             }
             c => Err(de::Error::invalid_type(
                 de::Unexpected::Char(c),
                 &format!(
-                    "`{}`, `{}` or `{}`",
-                    VARIANT_SIGNATURE_CHAR, ARRAY_SIGNATURE_CHAR, STRUCT_SIG_START_CHAR,
+                    "`{VARIANT_SIGNATURE_CHAR}`, `{ARRAY_SIGNATURE_CHAR}` or `{STRUCT_SIG_START_CHAR}`",
                 )
                 .as_str(),
             )),
@@ -320,18 +362,41 @@ where
     where
         V: Visitor<'de>,
     {
-        visitor.visit_enum(crate::de::Enum {
-            de: self,
+        let signature = self.0.sig_parser.next_signature()?;
+        let alignment = alignment_for_signature(&signature, self.0.ctxt.format())?;
+        self.0.parse_padding(alignment)?;
+
+        let non_unit = if self.0.sig_parser.next_char()? == STRUCT_SIG_START_CHAR {
+            // This means we've a non-unit enum. Let's skip the `(`.
+            self.0.sig_parser.skip_char()?;
+
+            true
+        } else {
+            false
+        };
+
+        let v = visitor.visit_enum(crate::de::Enum {
+            de: &mut *self,
             name,
-            phantom: PhantomData,
-        })
+            _phantom: PhantomData,
+        })?;
+
+        if non_unit {
+            // For non-unit enum, we need to skip the closing paren.
+            self.0.sig_parser.skip_char()?;
+        }
+
+        Ok(v)
+    }
+
+    fn is_human_readable(&self) -> bool {
+        false
     }
 }
 
-fn deserialize_ay<'de, 'sig, 'f, B>(de: &mut Deserializer<'de, 'sig, 'f, B>) -> Result<&'de [u8]>
-where
-    B: byteorder::ByteOrder,
-{
+fn deserialize_ay<'de, #[cfg(unix)] F: AsFd, #[cfg(not(unix))] F>(
+    de: &mut Deserializer<'de, '_, '_, F>,
+) -> Result<&'de [u8]> {
     if de.0.sig_parser.next_signature()? != "ay" {
         return Err(de::Error::invalid_type(de::Unexpected::Seq, &"ay"));
     }
@@ -342,8 +407,8 @@ where
     de.0.next_slice(len)
 }
 
-struct ArrayDeserializer<'d, 'de, 'sig, 'f, B> {
-    de: &'d mut Deserializer<'de, 'sig, 'f, B>,
+struct ArrayDeserializer<'d, 'de, 'sig, 'f, F> {
+    de: &'d mut Deserializer<'de, 'sig, 'f, F>,
     len: usize,
     start: usize,
     // alignment of element
@@ -358,18 +423,18 @@ struct ArrayDeserializer<'d, 'de, 'sig, 'f, B> {
     key_offset_size: Option<FramingOffsetSize>,
 }
 
-impl<'d, 'de, 'sig, 'f, B> ArrayDeserializer<'d, 'de, 'sig, 'f, B>
-where
-    B: byteorder::ByteOrder,
+impl<'d, 'de, 'sig, 'f, #[cfg(unix)] F: AsFd, #[cfg(not(unix))] F>
+    ArrayDeserializer<'d, 'de, 'sig, 'f, F>
 {
-    fn new(de: &'d mut Deserializer<'de, 'sig, 'f, B>) -> Result<Self> {
+    fn new(de: &'d mut Deserializer<'de, 'sig, 'f, F>) -> Result<Self> {
+        de.0.container_depths = de.0.container_depths.inc_array()?;
         let mut len = de.0.bytes.len() - de.0.pos;
 
         let element_signature = de.0.sig_parser.next_signature()?;
-        let element_alignment = alignment_for_signature(&element_signature, de.0.ctxt.format());
+        let element_alignment = alignment_for_signature(&element_signature, de.0.ctxt.format())?;
         let element_signature_len = element_signature.len();
         let fixed_sized_child = crate::utils::is_fixed_sized_signature(&element_signature)?;
-        let fixed_sized_key = if de.0.sig_parser.next_char() == DICT_ENTRY_SIG_START_CHAR {
+        let fixed_sized_key = if de.0.sig_parser.next_char()? == DICT_ENTRY_SIG_START_CHAR {
             // Key signature can only be 1 char
             let key_signature = Signature::from_str_unchecked(&element_signature[1..2]);
 
@@ -386,7 +451,7 @@ where
 
         let (offsets, offsets_len, key_offset_size) = if !fixed_sized_child {
             let (array_offsets, offsets_len) =
-                FramingOffsets::from_encoded_array(&de.0.bytes[de.0.pos..]);
+                FramingOffsets::from_encoded_array(subslice(de.0.bytes, de.0.pos..)?)?;
             len -= offsets_len;
             let key_offset_size = if !fixed_sized_key {
                 // The actual offset for keys is calculated per key later, this is just to
@@ -403,7 +468,7 @@ where
         };
         let start = de.0.pos;
 
-        if de.0.sig_parser.next_char() == DICT_ENTRY_SIG_START_CHAR {
+        if de.0.sig_parser.next_char()? == DICT_ENTRY_SIG_START_CHAR {
             de.0.sig_parser.skip_char()?;
         }
 
@@ -422,7 +487,7 @@ where
     fn element_end(&mut self, pop: bool) -> Result<usize> {
         match self.offsets.as_mut() {
             Some(offsets) => {
-                assert_eq!(self.de.0.ctxt.format(), EncodingFormat::GVariant);
+                assert_eq!(self.de.0.ctxt.format(), Format::GVariant);
 
                 let offset = if pop { offsets.pop() } else { offsets.peek() };
                 match offset {
@@ -443,9 +508,8 @@ where
     }
 }
 
-impl<'d, 'de, 'sig, 'f, B> SeqAccess<'de> for ArrayDeserializer<'d, 'de, 'sig, 'f, B>
-where
-    B: byteorder::ByteOrder,
+impl<'d, 'de, 'sig, 'f, #[cfg(unix)] F: AsFd, #[cfg(not(unix))] F> SeqAccess<'de>
+    for ArrayDeserializer<'d, 'de, 'sig, 'f, F>
 {
     type Error = Error;
 
@@ -459,27 +523,30 @@ where
                 .sig_parser
                 .skip_chars(self.element_signature_len)?;
             self.de.0.pos += self.offsets_len;
+            self.de.0.container_depths = self.de.0.container_depths.dec_array();
 
             return Ok(None);
         }
 
-        let ctxt = EncodingContext::new(
+        let ctxt = Context::new(
             self.de.0.ctxt.format(),
+            self.de.0.ctxt.endian(),
             self.de.0.ctxt.position() + self.de.0.pos,
         );
         let end = self.element_end(true)?;
 
-        let mut de = Deserializer::<B>(crate::DeserializerCommon {
+        let mut de = Deserializer::<F>(DeserializerCommon {
             ctxt,
             sig_parser: self.de.0.sig_parser.clone(),
-            bytes: &self.de.0.bytes[self.de.0.pos..end],
+            bytes: subslice(self.de.0.bytes, self.de.0.pos..end)?,
             fds: self.de.0.fds,
             pos: 0,
-            b: PhantomData,
+            container_depths: self.de.0.container_depths,
         });
 
         let v = seed.deserialize(&mut de).map(Some);
         self.de.0.pos += de.0.pos;
+        // No need for retaking the container depths as the child can't be incomplete.
 
         if self.de.0.pos > self.start + self.len {
             return Err(serde::de::Error::invalid_length(
@@ -492,9 +559,8 @@ where
     }
 }
 
-impl<'d, 'de, 'sig, 'f, B> MapAccess<'de> for ArrayDeserializer<'d, 'de, 'sig, 'f, B>
-where
-    B: byteorder::ByteOrder,
+impl<'d, 'de, 'sig, 'f, #[cfg(unix)] F: AsFd, #[cfg(not(unix))] F> MapAccess<'de>
+    for ArrayDeserializer<'d, 'de, 'sig, 'f, F>
 {
     type Error = Error;
 
@@ -515,8 +581,9 @@ where
 
         self.de.0.parse_padding(self.element_alignment)?;
 
-        let ctxt = EncodingContext::new(
+        let ctxt = Context::new(
             self.de.0.ctxt.format(),
+            self.de.0.ctxt.endian(),
             self.de.0.ctxt.position() + self.de.0.pos,
         );
         let element_end = self.element_end(false)?;
@@ -534,16 +601,17 @@ where
             None => element_end,
         };
 
-        let mut de = Deserializer::<B>(crate::DeserializerCommon {
+        let mut de = Deserializer::<F>(DeserializerCommon {
             ctxt,
             sig_parser: self.de.0.sig_parser.clone(),
-            bytes: &self.de.0.bytes[self.de.0.pos..key_end],
+            bytes: subslice(self.de.0.bytes, self.de.0.pos..key_end)?,
             fds: self.de.0.fds,
             pos: 0,
-            b: PhantomData,
+            container_depths: self.de.0.container_depths,
         });
         let v = seed.deserialize(&mut de).map(Some);
         self.de.0.pos += de.0.pos;
+        // No need for retaking the container depths as the key can't be incomplete.
 
         if self.de.0.pos > self.start + self.len {
             return Err(serde::de::Error::invalid_length(
@@ -559,8 +627,9 @@ where
     where
         V: DeserializeSeed<'de>,
     {
-        let ctxt = EncodingContext::new(
+        let ctxt = Context::new(
             self.de.0.ctxt.format(),
+            self.de.0.ctxt.endian(),
             self.de.0.ctxt.position() + self.de.0.pos,
         );
         let element_end = self.element_end(true)?;
@@ -572,16 +641,17 @@ where
         // Skip key signature (always 1 char)
         sig_parser.skip_char()?;
 
-        let mut de = Deserializer::<B>(crate::DeserializerCommon {
+        let mut de = Deserializer::<F>(DeserializerCommon {
             ctxt,
             sig_parser,
-            bytes: &self.de.0.bytes[self.de.0.pos..value_end],
+            bytes: subslice(self.de.0.bytes, self.de.0.pos..value_end)?,
             fds: self.de.0.fds,
             pos: 0,
-            b: PhantomData,
+            container_depths: self.de.0.container_depths,
         });
         let v = seed.deserialize(&mut de);
         self.de.0.pos += de.0.pos;
+        // No need for retaking the container depths as the value can't be incomplete.
 
         if let Some(key_offset_size) = self.key_offset_size {
             self.de.0.pos += key_offset_size as usize;
@@ -599,8 +669,8 @@ where
 }
 
 #[derive(Debug)]
-struct StructureDeserializer<'d, 'de, 'sig, 'f, B> {
-    de: &'d mut Deserializer<'de, 'sig, 'f, B>,
+struct StructureDeserializer<'d, 'de, 'sig, 'f, F> {
+    de: &'d mut Deserializer<'de, 'sig, 'f, F>,
     start: usize,
     end: usize,
     // Length of all the offsets after the array
@@ -609,9 +679,8 @@ struct StructureDeserializer<'d, 'de, 'sig, 'f, B> {
     offset_size: FramingOffsetSize,
 }
 
-impl<'d, 'de, 'sig, 'f, B> SeqAccess<'de> for StructureDeserializer<'d, 'de, 'sig, 'f, B>
-where
-    B: byteorder::ByteOrder,
+impl<'d, 'de, 'sig, 'f, #[cfg(unix)] F: AsFd, #[cfg(not(unix))] F> SeqAccess<'de>
+    for StructureDeserializer<'d, 'de, 'sig, 'f, F>
 {
     type Error = Error;
 
@@ -619,8 +688,9 @@ where
     where
         T: DeserializeSeed<'de>,
     {
-        let ctxt = EncodingContext::new(
+        let ctxt = Context::new(
             self.de.0.ctxt.format(),
+            self.de.0.ctxt.endian(),
             self.de.0.ctxt.position() + self.de.0.pos,
         );
         let element_signature = self.de.0.sig_parser.next_signature()?;
@@ -628,17 +698,25 @@ where
         let element_end = if !fixed_sized_element {
             let next_sig_pos = element_signature.len();
             let parser = self.de.0.sig_parser.slice(next_sig_pos..);
-            if !parser.done() && parser.next_char() == STRUCT_SIG_END_CHAR {
+            if !parser.done() && parser.next_char()? == STRUCT_SIG_END_CHAR {
                 // This is the last item then and in GVariant format, we don't have offset for it
                 // even if it's non-fixed-sized.
                 self.end
             } else {
                 let end = self
                     .offset_size
-                    .read_last_offset_from_buffer(&self.de.0.bytes[self.start..self.end])
+                    .read_last_offset_from_buffer(subslice(self.de.0.bytes, self.start..self.end)?)
                     + self.start;
-                self.end -= self.offset_size as usize;
-                self.offsets_len += self.offset_size as usize;
+                let offset_size = self.offset_size as usize;
+                if offset_size > self.end {
+                    return Err(serde::de::Error::invalid_length(
+                        offset_size,
+                        &format!("< {}", self.end).as_str(),
+                    ));
+                }
+
+                self.end -= offset_size;
+                self.offsets_len += offset_size;
 
                 end
             }
@@ -647,18 +725,19 @@ where
         };
 
         let sig_parser = self.de.0.sig_parser.clone();
-        let mut de = Deserializer::<B>(crate::DeserializerCommon {
+        let mut de = Deserializer::<F>(DeserializerCommon {
             ctxt,
             sig_parser,
-            bytes: &self.de.0.bytes[self.de.0.pos..element_end],
+            bytes: subslice(self.de.0.bytes, self.de.0.pos..element_end)?,
             fds: self.de.0.fds,
             pos: 0,
-            b: PhantomData,
+            container_depths: self.de.0.container_depths,
         });
         let v = seed.deserialize(&mut de).map(Some);
         self.de.0.pos += de.0.pos;
+        // No need for retaking the container depths as the field can't be incomplete.
 
-        if de.0.sig_parser.next_char() == STRUCT_SIG_END_CHAR {
+        if de.0.sig_parser.next_char()? == STRUCT_SIG_END_CHAR {
             // Last item in the struct
             de.0.sig_parser.skip_char()?;
 
@@ -673,8 +752,8 @@ where
 }
 
 #[derive(Debug)]
-struct ValueDeserializer<'d, 'de, 'sig, 'f, B> {
-    de: &'d mut Deserializer<'de, 'sig, 'f, B>,
+struct ValueDeserializer<'d, 'de, 'sig, 'f, F> {
+    de: &'d mut Deserializer<'de, 'sig, 'f, F>,
     stage: ValueParseStage,
     sig_start: usize,
     sig_end: usize,
@@ -682,14 +761,20 @@ struct ValueDeserializer<'d, 'de, 'sig, 'f, B> {
     value_end: usize,
 }
 
-impl<'d, 'de, 'sig, 'f, B> ValueDeserializer<'d, 'de, 'sig, 'f, B>
-where
-    B: byteorder::ByteOrder,
+impl<'d, 'de, 'sig, 'f, #[cfg(unix)] F: AsFd, #[cfg(not(unix))] F>
+    ValueDeserializer<'d, 'de, 'sig, 'f, F>
 {
-    fn new(de: &'d mut Deserializer<'de, 'sig, 'f, B>) -> Result<Self> {
+    fn new(de: &'d mut Deserializer<'de, 'sig, 'f, F>) -> Result<Self> {
         // GVariant format has signature at the end
         let mut separator_pos = None;
 
+        if de.0.bytes.is_empty() {
+            return Err(de::Error::invalid_value(
+                de::Unexpected::Other("end of byte stream"),
+                &"nul byte separator between Variant's value & signature",
+            ));
+        }
+
         // Search for the nul byte separator
         for i in (de.0.pos..de.0.bytes.len() - 1).rev() {
             if de.0.bytes[i] == b'\0' {
@@ -709,7 +794,7 @@ where
             Some(separator_pos) => (separator_pos + 1, de.0.bytes.len(), de.0.pos, separator_pos),
         };
 
-        Ok(ValueDeserializer::<B> {
+        Ok(ValueDeserializer::<F> {
             de,
             stage: ValueParseStage::Signature,
             sig_start,
@@ -720,9 +805,8 @@ where
     }
 }
 
-impl<'d, 'de, 'sig, 'f, B> SeqAccess<'de> for ValueDeserializer<'d, 'de, 'sig, 'f, B>
-where
-    B: byteorder::ByteOrder,
+impl<'d, 'de, 'sig, 'f, #[cfg(unix)] F: AsFd, #[cfg(not(unix))] F> SeqAccess<'de>
+    for ValueDeserializer<'d, 'de, 'sig, 'f, F>
 {
     type Error = Error;
 
@@ -734,17 +818,17 @@ where
             ValueParseStage::Signature => {
                 self.stage = ValueParseStage::Value;
 
-                let signature = Signature::from_str_unchecked(VARIANT_SIGNATURE_STR);
+                let signature = Signature::from_static_str_unchecked(VARIANT_SIGNATURE_STR);
                 let sig_parser = SignatureParser::new(signature);
 
-                let mut de = Deserializer::<B>(crate::DeserializerCommon {
+                let mut de = Deserializer::<F>(DeserializerCommon {
                     // No padding in signatures so just pass the same context
                     ctxt: self.de.0.ctxt,
                     sig_parser,
-                    bytes: &self.de.0.bytes[self.sig_start..self.sig_end],
+                    bytes: subslice(self.de.0.bytes, self.sig_start..self.sig_end)?,
                     fds: self.de.0.fds,
                     pos: 0,
-                    b: PhantomData,
+                    container_depths: self.de.0.container_depths,
                 });
 
                 seed.deserialize(&mut de).map(Some)
@@ -752,22 +836,23 @@ where
             ValueParseStage::Value => {
                 self.stage = ValueParseStage::Done;
 
-                let slice = &self.de.0.bytes[self.sig_start..self.sig_end];
+                let slice = subslice(self.de.0.bytes, self.sig_start..self.sig_end)?;
                 // FIXME: Can we just use `Signature::from_bytes_unchecked`?
                 let signature = Signature::try_from(slice)?;
                 let sig_parser = SignatureParser::new(signature);
 
-                let ctxt = EncodingContext::new(
+                let ctxt = Context::new(
                     self.de.0.ctxt.format(),
+                    self.de.0.ctxt.endian(),
                     self.de.0.ctxt.position() + self.value_start,
                 );
-                let mut de = Deserializer::<B>(crate::DeserializerCommon {
+                let mut de = Deserializer::<F>(DeserializerCommon {
                     ctxt,
                     sig_parser,
-                    bytes: &self.de.0.bytes[self.value_start..self.value_end],
+                    bytes: subslice(self.de.0.bytes, self.value_start..self.value_end)?,
                     fds: self.de.0.fds,
                     pos: 0,
-                    b: PhantomData,
+                    container_depths: self.de.0.container_depths.inc_variant()?,
                 });
 
                 let v = seed.deserialize(&mut de).map(Some);
@@ -781,23 +866,8 @@ where
     }
 }
 
-impl<'de, 'd, 'sig, 'f, B> crate::de::GetDeserializeCommon<'de, 'sig, 'f, B>
-    for &'d mut Deserializer<'de, 'sig, 'f, B>
-where
-    B: byteorder::ByteOrder,
-{
-    fn common_mut<'dr>(self) -> &'dr mut crate::de::DeserializerCommon<'de, 'sig, 'f, B>
-    where
-        Self: 'dr,
-    {
-        &mut self.0
-    }
-}
-
-impl<'de, 'd, 'sig, 'f, B> EnumAccess<'de>
-    for crate::de::Enum<B, &'d mut Deserializer<'de, 'sig, 'f, B>>
-where
-    B: byteorder::ByteOrder,
+impl<'de, 'd, 'sig, 'f, #[cfg(unix)] F: AsFd, #[cfg(not(unix))] F> EnumAccess<'de>
+    for crate::de::Enum<&'d mut Deserializer<'de, 'sig, 'f, F>, F>
 {
     type Error = Error;
     type Variant = Self;
